/* -*- c++ -*-

  SigC++ cross-thread signalling support.
   
  Copyright 2000 Andreas Rottmannn <rottmann@users.sourceforge.net>
  Copyright 2004 Martin Schulze <mschulze@cvs.gnome.org>
                                                         
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.
 
  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <unistd.h> // for usleep()
#include "gtkmm_dispatcher2.h"


namespace sigc {

static void null_func()
  {}

GtkmmDispatcher::GtkmmDispatcher() :
  disp_thread_( Glib::Thread::self() ), notifier_( Glib::MainContext::get_default() ), abort_(false) {

  notifier_.connect(sigc::mem_fun(*this, &GtkmmDispatcher::do_slots));
}

GtkmmDispatcher::GtkmmDispatcher( const Glib::RefPtr<Glib::MainContext>& context ) :
  disp_thread_( Glib::Thread::self() ), notifier_( context ), abort_(false) {

  notifier_.connect(sigc::mem_fun(*this, &GtkmmDispatcher::do_slots));
}

GtkmmDispatcher::~GtkmmDispatcher() {
  // Synchronize, so all previous requests get done
  if (!abort_) tunnel(sigc::ptr_fun(&null_func), true);
}

void GtkmmDispatcher::tunnel(const sigc::slot<void>& slot_, bool sync) {
  if( sync && ( Glib::Thread::self() == disp_thread_ ) ) {
    // We have the dispatcher in the same thread
    slot_();
  } else {
    disp_queue_push(slot_);
    if (sync) {
      Sync sync_;
      sync_.mutex.lock();
      disp_queue_push(sigc::mem_fun(sync_, &Sync::signal));
      sync_.cond.wait(sync_.mutex);
      sync_.mutex.unlock();
    }
  }
}

void GtkmmDispatcher::do_slots() {
  std::pair<sigc::slot<void>,bool> vslot;

 try {
  while (has_work() && !abort_) {
    vslot = disp_queue_.pop();
    if (vslot.second) vslot.first();
  }
 } catch( ... ) {
 }
}

void GtkmmDispatcher::resume() {
  notifier_.emit();
}

void GtkmmDispatcher::abort() {
  abort_ = true;
  notifier_.emit();
}

void GtkmmDispatcher::disp_queue_push(const sigc::slot<void> s) {
  while (!disp_queue_.push(s)) usleep(100);
  resume();
}


} /* namespace sigc */
