/* -*- c++ -*-

  SigC++ cross-thread signalling support.
   
  Copyright 2000 Andreas Rottmannn <rottmann@users.sourceforge.net>
  Copyright 2004 Martin Schulze <mschulze@cvs.gnome.org>
                                                         
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.
 
  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef SIGCXX_GTKMM_DISPATCHER_H
#define SIGCXX_GTKMM_DISPATCHER_H

#include <stdexcept>
#include <list>

#include <sigc++/functors/slot.h>
#include <glibmm/main.h>
#include <glibmm/thread.h>
#include <glibmm/dispatcher.h>
#include "ringbuffer.h"


namespace sigc {

class Dispatcher {
  public:
    virtual ~Dispatcher() {}
    virtual void tunnel(const sigc::slot<void>& slot_, bool sync = false) = 0;
};

class GtkmmDispatcher : public Dispatcher {
public:
  GtkmmDispatcher();
  GtkmmDispatcher(const Glib::RefPtr<Glib::MainContext>& context);
  ~GtkmmDispatcher();

  void set_thread(Glib::Thread* thread) { disp_thread_ = thread; }

  virtual void tunnel(const sigc::slot<void>& slot_, bool sync = false);

  void do_slots();
  bool has_work() const { return disp_queue_.size() > 0; }

//  void msleep( unsigned short msec = 0 ); // may only be called by the Thread itself !!!
     // wait for some other thread to call resume() if should_quit is false
     // if msec > 0 wait times out after msec milliseconds / otherwise infinite wait if has_work returns false
  void resume();
  void abort();

private:
  struct Sync {
    Glib::Mutex mutex;
    Glib::Cond  cond;
    void signal()
      { Glib::Mutex::Lock lock(mutex); cond.signal(); }
  };

  Glib::Thread* disp_thread_;
  Util::ringbuffer<sigc::slot<void>,500> disp_queue_;
  void disp_queue_push(const sigc::slot<void> s);

  Glib::Dispatcher notifier_;
  bool abort_;
};

} /* namespace sigc */
 

#endif
