/***************************************************************************
                          ringbuffer.cpp  -  description
                             -------------------
    begin                : Tue Apr 17 2001
    copyright            : (C) 2001 by Martin Schulze
    email                : martin-ml@hippogriff.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <string.h>
#include <assert.h>
#include "ringbuffer.h"



namespace Util {


ringbuffer_::ringbuffer_( ringbuffer_::BYTE* vchilds, unsigned short vchldSize,
  unsigned short vrealSize ) : childs( vchilds ), appendMask( 0 ), removeMask( 0 ),
  firstPos( 0 ), lastPos( 0 ), size_( 0 ),
  chldSize( vchldSize ), realSize( vrealSize ) {

  if( realSize == USHRT_MAX ) throw "E_Memory"; // ThrowException( EX_Memory ); TODO

  unsigned short tempSize = (unsigned short)( (double)realSize / 8 + 0.99999 );
  appendMask = new BYTE [ tempSize ];
  removeMask = new BYTE [ tempSize ];
  memset( appendMask, 255, tempSize );
  memset( removeMask, 0, tempSize );
}

ringbuffer_::~ringbuffer_() {
  if( appendMask ) {
    delete [] appendMask;
    delete [] removeMask;
  }
}

unsigned short ringbuffer_::push() {
  unsigned short pos = last();
  unsigned short stopPos = pos;

  if( full() )
    return USHRT_MAX;

  while( !test_and_clear_bit( pos, appendMask ) )
    if( incr( pos ) == stopPos )
      return USHRT_MAX;

  lastPos++;
  if( lastPos >= realSize ) lastPos -= realSize; // kann mehrfach ausgefuehrt werden
  size_++;

  return pos;
}

void ringbuffer_::push_finish( unsigned short pos ) {
  assert( pos < realSize );
  set_bit( pos, removeMask );
}

unsigned short ringbuffer_::pop() {
  unsigned short pos = first();
  unsigned short stopPos = pos;

  if( empty() )
    return USHRT_MAX;

  while( !test_and_clear_bit( pos, removeMask ) )
    if( incr( pos ) == stopPos )
      return USHRT_MAX;

  firstPos++;
  if( firstPos >= realSize ) firstPos -= realSize; // kann mehrfach ausgefuehrt werden
  size_--;

  return pos;
}

void ringbuffer_::pop_finish( unsigned short pos ) {
  assert( pos < realSize );
  set_bit( pos, appendMask );
}

void ringbuffer_::clear() {
  size_ = 0;
  firstPos = lastPos = 0;

  unsigned short tempSize = (unsigned short)( (double)realSize / 8 + 0.99999 );
  memset( appendMask, 255, tempSize );
  memset( removeMask, 0, tempSize );
}


}
